/**
 * Defines for how long DeviceProxy will keep responses
 *
 * @type {number}
 */
mtango.EventTypeEnum = Object.freeze({
    CHANGE:"change",
    PERIODIC:"periodic",
    ARCHIVE:"archive",
    USER:"user"
});
/**
 * Model
 *
 * @type {mtango.DeviceAttribute}
 */
mtango.DeviceAttribute = MVC.Model.extend('DeviceAttribute',
    /* @Static */
    {
        _ID: 1,//global counter
        attributes: {
            name : "string",
            label: "string",
            description: "string",
            isReadOnly: "boolean",
            dataFormat: "string",
            dataType: "string",
            unit    : "string",
            displayUnit: "string",
            maxDimX: "int",
            maxDimY: "int",
            maxValue: "string",
            minValue: "string"
        },
        default_attributes: {},
        /**
         * Creates a new instance of DeviceAttribute and populates it's properties (name, label etc) using network call
         *
         * @param {String} device
         * @param {String} attr_name
         * @param {Object} params callbacks
         * @asynchronous
         */
        create:function(device, attr_name, params) {
            var callbacks = this._clean_callbacks(params);

            var proxy = mtango.DeviceProxy.find(device);
            if (proxy == null) throw "Can not create DeviceAttribute! No proxy was found for device: " + device;
            new MVC.JsonP(proxy.full_url + attr_name + "/info", {
                onComplete: MVC.Function.bind(function (response) {
                    var attributes = {
                        id: this._ID++,
                        proxy: proxy
                    };
                    MVC.Object.extend(attributes, response);
                    var instance = this.create_as_existing(attributes);
                    if (instance.valid())
                        callbacks.onComplete(instance);
                    else
                        callbacks.onFailure(instance);
                }, this)
            });
        }
    },
    /* @Prototype */
    {
        timeout:5000,
        proxy: null,
        response: null,
        _subscriptions: null,
        /**
         * Sets a proxy backend for this attribute
         *
         * @param v
         */
        set_proxy:function(v) {
            this.proxy = v;
        },
        /**
         *
         * @param {mtango.Response} v
         */
        set_response:function(v){
            this.value = v;
        },
        /**
         *
         * @param {int} v
         */
        set_timeout:function(v){
            this.timeout = v;
        },
        /**
         * Returns the suggested element id for this instance
         */
        element_id:function(){
            return this.proxy.device+'/'+this.name+'_'+this.id;
        },
        /**
         *
         * @param attributes
         * @constructor
         */
        init:function(attributes){
            if(!attributes.id) attributes.id = this.Class._ID++;
            this._super(attributes);

            this._subscriptions = {
                change:[],
                periodic:[],
                archive:[],
                user:[]
            };
        },
        /**
         * Reads this attribute value through the proxy
         *
         * @param cbk
         * @asynchronous
         */
        read:function(cbk) {
            this.proxy.readAttribute(this.name, {timeout:this.timeout}, cbk);
        },
        /**
         * Writes a new value to this attribute through the proxy
         *
         * @param {*} argin
         * @param {Object|Function} cbk
         * @asynchronous
         */
        write:function(argin, cbk) {
            if (this.isReadOnly) return;
            var params = {
                argin: argin,
                timeout:this.timeout
            };
            this.proxy.writeAttribute(this.name, params, cbk);
        },
        poll_id: 0,
        is_polling: false,
        /**
         * Starts continuous polling
         *
         * @param cbk
         * @param interval
         */
        start_polling:function(cbk, interval) {
            console.debug("Starting polling of " + this.element_id());
            this.poll_id = setInterval(
            MVC.Function.bind(function () {
                if (this.is_polling)
                    this.read(cbk)
                }, this),
                interval);
            this.is_polling = true;
        },
        /**
         * Pauses polling
         */
        pause_polling:function(){
            console.debug("Pausing polling of " + this.element_id());
            this.is_polling = false;
        },
        /**
         * Resumes polling
         */
        resume_polling:function(){
            console.debug("Resuming polling of " + this.element_id());
            this.is_polling = true;
        },
        /**
         * Cancels polling. After this method to resume polling one must call start_polling
         */
        stop_polling:function() {
            console.debug("Canceling polling of " + this.element_id());
            clearInterval(this.poll_id);
            this.is_polling = false;
        },
        /**
         * Subscribes to change event of this attribute
         *
         * @param {Function} cbk onEvent
         * @asynchronous
         */
        change:function(cbk){
            var id = this.proxy.subscribe(this.name,mtango.EventTypeEnum.CHANGE,cbk,this.timeout);
            this._subscriptions.change.push(id);
        },
        /**
         * Subscribes to periodic event of this attribute
         *
         * @param {Function} cbk onEvent
         * @asynchronous
         */
        periodic:function(cbk){
            var id = this.proxy.subscribe(this.name,mtango.EventTypeEnum.PERIODIC,cbk,this.timeout);
            this._subscriptions.periodic.push(id);
        },
        /**
         * Subscribes to archive event of this attribute
         *
         * @param {Function} cbk onEvent
         * @asynchronous
         */
        archive:function(cbk){
            var id = this.proxy.subscribe(this.name,mtango.EventTypeEnum.ARCHIVE,cbk,this.timeout);
            this._subscriptions.archive.push(id);
        },
        /**
         * Subscribes to user event of this attribute
         *
         * @param {Function} cbk onEvent
         * @asynchronous
         */
        user:function(cbk){
            var id = this.proxy.subscribe(this.name,mtango.EventTypeEnum.USER,cbk,this.timeout);
            this._subscriptions.user.push(id);
        },
        /**
         * Cancels subscription to change event of this attribute
         */
        cancel_change:function(){
            var me = this;
            this._subscriptions.change.forEach(
                function(id){
                    me.proxy.unsubscribe(me.name,mtango.EventTypeEnum.CHANGE,id);
                }
            );
        },
        /**
         * Cancels subscription to periodic event of this attribute
         */
        cancel_periodic:function(){
            var me = this;
            this._subscriptions.periodic.forEach(
                function(id){
                    me.proxy.unsubscribe(me.name,mtango.EventTypeEnum.CHANGE,id);
                }
            );
        },
        /**
         * Cancels subscription to archive event of this attribute
         */
        cancel_archive:function(){
            var me = this;
            this._subscriptions.archive.forEach(
                function(id){
                    me.proxy.unsubscribe(me.name,mtango.EventTypeEnum.CHANGE,id);
                }
            );
        },
        /**
         * Cancels subscription to user event of this attribute
         */
        cancel_user:function(){
            var me = this;
            this._subscriptions.user.forEach(
                function(id){
                    me.proxy.unsubscribe(me.name,mtango.EventTypeEnum.CHANGE,id);
                }
            );
        },
        /**
         * Cancels all subscriptions of this DeviceAttribute instance
         */
        cancel_all:function(){
            for(var evt in this._subscriptions){
                if(!this._subscriptions.hasOwnProperty(evt)) continue;
                var me = this;
                this._subscriptions[evt].forEach(
                    function(id){
                        me.proxy.unsubscribe(me.name,evt,id);
                    }
                );
            }
        },
        /**
         * {device}/{this.name}, i.e. sys/tg_test/0/double_scalar
         *
         * @returns {string}
         */
        toString:function(){
            return this.proxy.device + '/' + this.name;
        }
    }
);

if(typeof EventTypeEnum == 'undefined') EventTypeEnum = mtango.EventTypeEnum;
if(typeof DeviceAttribute == 'undefined') DeviceAttribute = mtango.DeviceAttribute;