/**
 * Represents device's command.
 *
 *
 * @type {mtango.DeviceCommand}
 */
mtango.DeviceCommand = MVC.Model.extend('DeviceCommand',
    /* @Static */
    {
        _ID : 1,
        attributes:{
            name: 'string',
            inputType   : 'string',
            inputDescription : 'string',
            outputType  : 'string',
            outputDescription : 'string'
        },
        default_attributes:{
            id: 0
        },
        /**
         * Creates an instance of this model.
         *
         * A JsonP call is executed to obtain information about the command from the remote tango device
         *
         * @param {string} device -- e.g. 'sys/tg_test/1'
         * @param {string} cmd_name -- command's name
         * @param {Object|Function} cbks -- callback(s)
         * @asynchronous
         */
        create:function(device, cmd_name, cbks){
            var callbacks = this._clean_callbacks(cbks);

            var proxy = mtango.DeviceProxy.find(device);
            if (proxy == null) throw "Can not create DeviceCommand! No proxy was found for device: " + device;
            new MVC.JsonP(proxy.full_url + cmd_name + "/info", {
                onComplete: MVC.Function.bind(function (response) {
                    var attributes = {
                        id: this._ID++,
                        proxy: proxy
                    };
                    MVC.Object.extend(attributes, response);
                    var instance = this.create_as_existing(attributes);
                    if (instance.valid())
                        callbacks.onComplete(instance);
                    else
                        callbacks.onFailure(instance);
                }, this)
            });
        }
    },
    /* @Prototype */
    {
        timeout : 5000,
        proxy: null,
        response : null,
        /**
         * Sets this command request timeout
         *
         * @param {int} v
         */
        set_timeout:function(v){
            this.timeout = v;
        },
        /**
         * Sets this command proxy
         *
         * @param {mtango.DeviceProxy} v
         */
        set_proxy:function(v){
            this.proxy = v;
        },
        /**
         * Sets this command response
         *
         * @param {mtango.Response} v
         */
        set_response:function(v){
            this.response = v;
        },
        /**
         * Returns this command element id, i.e. sys/tg_test/1/DevString_1
         *
         * @returns {string}
         */
        element_id:function(){
            return this.proxy.device+'/'+this.name+'_'+this.id;
        },
        /**
         *
         * @param attributes
         * @constructor
         */
        init:function(attributes){
            if(!attributes.id) attributes.id = this.Class._ID++;
            this._super(attributes);
        },
        /**
         * Executes this command on the remote tango device
         *
         * @param {*} argin
         * @param {Object|Function} cbk -- callback(s)
         */
        execute:function(argin, cbk){
            if(typeof argin == 'function' || argin == 'Object') {
                cbk = argin;
                argin = null;
            }
            this.proxy.executeCommand(this.name, {argin:argin},cbk);
        },
        /**
         * @return {string}
         */
        toString:function(){
            return this.proxy.device + '/' + this.name;
        }
    }
);

if(typeof DeviceCommand == 'undefined') DeviceCommand = mtango.DeviceCommand;