
/**
 * Provides similar functionality as org.tango.client.DeviceProxy from TangORB
 *
 * @type {mtango.DeviceProxy}
 */
mtango.DeviceProxy = MVC.Model.extend('DeviceProxy',
/* @Static */
{
    /**
     * For internal use only
     */
    _stateEnum : Object.freeze({
        UNDEFINED   : 'UNDEFINED',
        INITIAL     : 'INITIAL',
        CONTINUATION: 'CONTINUATION'
    }),
    kDeviceProxyDelay:200,
    kDeviceProxyRequestValidTime:200,
    api:'mtango',
    id:'device',
    error_timeout: 5000, // 5s
    attributes:{
        __requests:'Object',
        __subscriptions:'Object',
        __pendingRequests:'Object'//url -> [awaiting requests]
    },
    default_attributes:{//this will be shared across all instances
        __requests:{},
        __subscriptions:{},
        __pendingRequests:{}
    },
    _subscriptionId : 1,
    //create URL
    //init method
    /**
     *
     * @param url
     * @param cmd
     * @param params
     * @param cbs
     * @private
     */
    _sendRequest:function(url, params, cbs){
        console.debug("Sending request: " + url);

        var onSuccess = function(response){
                response.src = url;
                response = new mtango.Response(response);
                if(response.valid()) {
                    console.debug('Request has succeed: ' + url);
                    cbs.onSuccess(response);
                }
                else {
                    console.debug('Request has failed:' + url);
                    response.errors.forEach(function(error){
                        console.error(error.severity + ": " + error.reason + " - " + error.description)
                    });
                    cbs.onFailure(response);
                }
            };
        var onFailure = function(response){
            onSuccess({
                errors:[{
                    reason: "Timeout",
                    severity: "ERR",
                    description: response.responseText,
                    origin: "DeviceProxy.js#_sendRequest"
                }],
                quality:mtango.AttrQualityEnum.INVALID,
                timestamp:Date.now()
            })
        };

        var method = params.method || 'get';
        var timeout = params.timeout || this.error_timeout;


        new MVC.JsonP(url, {
            error_timeout : timeout,
            parameters:params,
            onSuccess :onSuccess,
            onFailure :onFailure,
            method    :method
        });
    },
    /**
     * Resends after getting response
     *
     * @param {string} url
     * @param {Object} params
     * @param {Function} onComplete
     * @private
     */
    _sendRequestContinuously:function(url, params, onComplete){
        var callbacks =  MVC.Model._clean_callbacks(
            MVC.Function.bind(
            function(response) {
                if (onComplete(response))
                    setTimeout(
                        MVC.Function.bind(
                            function(){
                                params.state = this._stateEnum.CONTINUATION;
                                this._sendRequest(url, params, callbacks);
                            }
                        ,this),this.kDeviceProxyDelay);
            },this)
        );
        params.state = this._stateEnum.INITIAL;
        this._sendRequest(url,params, callbacks);
    }

},
/* @Prototype */
{
    device:null,
    url:null,
    full_url:null,
    wrapped_by:null,
    set_url:function(v){
        this.url = v;
    },
    set_wrapped_by:function(v){
        this.wrapped_by = v;
    },
    /**
     *
     * @param attributes {url,device}
     * @constructor
     */
    init: function(attributes){
        this._super(attributes);
        this.full_url = this.url + "/" + this.Class.api + "/device/" + this.device + "/";
    },
    /**
     * Checks request status and acts accordingly
     *
     * Possible callbacks:
     * <ul>
     *      <li>params.onSuccess -- defines positive callback
     *      <li>params.onFailure -- defines negative callback
     * </ul>
     *
     * @param {string} url
     * @param {Action} cmd
     * @param {Object} params
     * @param {Object} cbs
     * @private
     */
    _sendRequest:function(url,cmd,params, cbs){
        var callbacks = MVC.Model._clean_callbacks(cbs);
        var onSuccess = callbacks.onSuccess;
        var onFailure  = callbacks.onFailure;

        var requestObj = this.__requests[url];
        if(!requestObj){
            //create new request object
            this.__requests[url] = new mtango.Request({status:RequestStateEnum.PENDING,cmd:cmd,result:null});

            //redefine callbacks
            callbacks.onSuccess = MVC.Function.bind(function(response){
                //set requestObj status
                this.__requests[url].result = response;
                this.__requests[url].status = RequestStateEnum.FINISHED;
                //remove cache after a while
                setTimeout(MVC.Function.bind(function(){
                    delete this.__requests[url];
                },this),this.Class.kDeviceProxyRequestValidTime);
                //invoke requests cbks
                if(response.valid()) {
                    OpenAjax.hub.publish(this.device + '/' + cmd.target + '.' + cmd.type + '.succeed', response);
                    onSuccess(response);
                } else {
                    OpenAjax.hub.publish(this.device + '/' + cmd.target + '.' + cmd.type + '.failed', response);
                    onFailure(response);
                }
                //invoke awaiting requests
                if(this.__pendingRequests[url] && this.__pendingRequests[url].length > 0) {
                    do{
                        this.__pendingRequests[url].pop()(response);
                    } while(this.__pendingRequests[url].length > 0);
                }
            },this);
            callbacks.onFailure = MVC.Function.bind(function(response){
                //set requestObj status
                this.__requests[url].result = response;
                this.__requests[url].status = RequestStateEnum.ERROR;
                //remove cache after a while
                setTimeout(MVC.Function.bind(function(){
                    delete this.__requests[url];
                },this),this.Class.kDeviceProxyRequestValidTime);
                //requestObj has failed
                OpenAjax.hub.publish(this.device+'/'+cmd.target+'.'+cmd.type+'.failed',response);
                onFailure(response);
                //invoke awaiting requests
                if(this.__pendingRequests[url] && this.__pendingRequests[url].length > 0) {
                    do{
                        this.__pendingRequests[url].pop()(response);
                    } while(this.__pendingRequests[url].length > 0);
                }
            },this);

            //send
            this.Class._sendRequest(url, params, callbacks);
        } else if(requestObj.isPending()){
            if(!this.__pendingRequests[url]) this.__pendingRequests[url] = [];
            this.__pendingRequests[url].push(
                function(response){
                    if(response.errors)
                        onFailure(response);
                    else
                        onSuccess(response)
                });
        } else if(requestObj.isFinished()){
            onSuccess(requestObj.result);
        } else if(requestObj.isError()){
            onFailure(requestObj.result);
        } else {
            console.error("Unreachable statement!");
            throw "Unreachable statement!"
        }
    },
    /**
     * Reads an attribute from the remote tango device
     *
     * @param {string} attributeName
     * @param {Object} params
     * @param {Object|Function} cbks -- callback(s)
     * @asynchronous
     */
    readAttribute:function(attributeName, params, cbks){
        var cmd = Action.createReadAction(attributeName);

        this._sendRequest(this.full_url + cmd.target, cmd, params, cbks);
    },
    /**
     * Reads an attribute info from the remote server
     *
     * @param {string} attributeName
     * @param {Object} params
     * @param {Object|Function} cbks -- callback(s)
     * @asynchronous
     */
    readAttributeInfo:function(attributeName,params, cbks){
        var cmd = mtango.Action.createReadAction(attributeName);

        this._sendRequest(this.full_url + cmd.target + "/info",cmd,params,cbks);
    },
    /**
     * Writes an attribute to the remote tango device
     *
     * @param {string} attributeName
     * @param {Object} params
     * @param {Object|Function} cbks -- callback(s)
     * @asynchronous
     */
    writeAttribute:function(attributeName, params, cbks){
        var cmd = mtango.Action.createWriteAction(attributeName,params.argin);
        var arg = encodeURI(cmd.argin).replace(/\./g,"%2E");
        this._sendRequest(this.full_url + cmd.target + "=" + arg, cmd, MVC.Object.extend(params, {method:'put'}), cbks);
    },
    /**
     * Executes command on a remote tango device
     *
     * @param {string} commandName
     * @param {Object} params
     * @param {Object|Function} cbks -- callback(s)
     * @asynchronous
     */
    executeCommand:function(commandName, params, cbks){
        var cmd = mtango.Action.createExecAction(commandName,params.argin || null);
        this._sendRequest(this.full_url + cmd.target + ((cmd.argin) ? "=" + encodeURI(cmd.argin).replace(/\./g,"%2E"): ""), cmd, params, cbks);
    },
    /**
     * Reads a command info from the remote tango device
     *
     * @param {string} commandName
     * @param {Object} params
     * @param {Object|Function} cbks -- callback(s)
     * @asynchronous
     */
    readCommandInfo:function(commandName, params, cbks){
        var cmd = mtango.Action.createReadAction(commandName);

        this._sendRequest(this.full_url + cmd.target + "/info",cmd,params,cbks);
    },
    /**
     * Subscribes this device proxy to a specified event
     *
     * @param attr
     * @param {mtango.EventTypeEnum} evt
     * @param {Function} handler
     * @param {int} timeout
     * @return {int} subscription id
     * @asynchronous
     */
    subscribe:function(attr, evt, handler, timeout){
        var subscription = this.device + "/" + attr + "." + evt;

        var id = this.Class._subscriptionId++;
        if(this.__subscriptions[subscription]){
            this.__subscriptions[subscription][id] = handler;
            this.__subscriptions[subscription].count++;
        } else {
            this.__subscriptions[subscription] = new Subscription(id,handler);
            this.Class._sendRequestContinuously(this.full_url + attr + "." + evt , {
                    timeout: timeout
                },
                MVC.Function.bind(function(response){
                    if(this.__subscriptions[subscription]){
                        for(var _id in this.__subscriptions[subscription]){
                            if(!this.__subscriptions[subscription].hasOwnProperty(_id) || _id == 'count') continue;
                            this.__subscriptions[subscription][_id](response);
                        }
                        return true;
                    } else
                        return false;
                },this)
            );
        }
        return id;
    },
    /**
     *
     * @param attr
     * @param {mtango.EventTypeEnum} evt
     * @param {int} id -- subscription id obtained in subscribe
     */
    unsubscribe:function(attr, evt, id){
        var subscription = (this.device + "/" + attr + "." + evt);

        if(typeof this.__subscriptions[subscription] == 'undefined') return;

        delete this.__subscriptions[subscription][id];
        this.__subscriptions[subscription].count--;
        if(this.__subscriptions[subscription].count == 0)
            delete this.__subscriptions[subscription];
    },
    /**
     * Shorthand to read state attribute from the device
     *
     * @param {Object} params
     * @param {Object|Function} cbks
     * @asynchronous
     */
    state:function(params, cbks){
        this.readAttribute('State', params, cbks);
    }
});

Subscription = function(id,handler){
    this.count = 1;
    this[id] = handler;
};

if(typeof DeviceProxy == 'undefined') DeviceProxy = mtango.DeviceProxy;