/**
 *
 * @type {MVC.Model}
 */
mtango.ui.DeviceAttributeUI = MVC.Model.extend("DeviceAttributeUI",
    {
        _ID : 1,
        kAcquisitionMethod: ['poll', 'read', 'change', 'periodic', 'archive', 'user'],
        kAvailableViews : {
            fld_text_ro:'engines/mTangoUI/views/fld_text_ro.ejs',
            fld_text_rw:'engines/mTangoUI/views/fld_text_rw.ejs',
            fld_list_ro:'engines/mTangoUI/views/fld_list_ro.ejs',
            fld_list_rw:'engines/mTangoUI/views/fld_list_rw.ejs',
            fld_empty  :'engines/mTangoUI/views/fld_empty.ejs'
        },
        attributes:{
            src:'string',
            view_url:'string',
            acquisition_method:'string',
            interval:'int',
            enhance:'boolean'
        },
        default_attributes:{
            id: 0,
            interval: 5000,
            enhance:false
        },
        _validator:
            function(inst){
                return function() {
                    if (!inst.src) inst.errors.push("Required attribute[src] is missing!");
                    if (!inst.view_url) inst.errors.push("Required attribute[view_url] is missing!");
                    if (!inst.acquisition_method) inst.errors.push("Required acquisition_method is missing. Specify one of the [" + inst.Class.kAcquisitionMethod + "]");
                    if (inst.acquisition_method != 'read' && !inst.interval) console.warn("No interval for non-read acquisition method: 5000 ms will be used by default.");
                }
            }
        ,
        /**
         *
         * @param $el a JQuery ref to DeviceAttributeUI element
         * @returns {String}
         * @private
         */
        _resolve_acquisition_method:function($el){
            for (var i = 0, size = this.kAcquisitionMethod.length; i < size; ++i) {
                var method = this.kAcquisitionMethod[i];
                if ($el.attr(method) !== undefined && $el.attr(method) !== false) return method;
            }
            return null;
        },
        /**
         *
         * @param v an url to .ejs file or a shortcut from this.kAvailableViews
         * @returns {MVC.View}
         * @private
         */
        _resole_view_url_to_view:function(v){
            return mtango.UIHelpers.resolve_view_url_to_view(v, this.kAvailableViews);
        },
        /**
         * Creates an instance of this model from wrapped, view_url and acquisition method
         *
         * @param {DeviceAttribute} deviceAttribute
         * @param {String} view_url
         * @param {String} acquisition_method
         * @param {int} interval (optional, used for polling)
         */
        create:function(deviceAttribute, view_url, acquisition_method, interval){
            if(!MVC.Array.include(this.kAcquisitionMethod,acquisition_method)) throw "Unknown acquisition_method: " + acquisition_method;

            view_url = this._check_readonly(deviceAttribute.isReadOnly, view_url);
            var view = this._resole_view_url_to_view(view_url);

            var instance = this.create_as_existing({
                src : deviceAttribute.toString(),
                view_url: view_url,
                view: view,
                acquisition_method:acquisition_method,
                interval          : interval || 0,
                wrapped:deviceAttribute
            });
            return instance;
        },
        /**
         * Inherited models may override this method to add extra attributes when parsing html
         *
         * @param {jQuery} $el
         * @returns {Object} parsed attributes
         * @private
         */
        _parse:function($el){
            console.debug("An attribute has been found on the page. Parsing ...");
            var result = {
                $element: $el,
                errors:[]
            };
            if($el.attr('id')) result.id = $el.attr('id');

            result.src = $el.attr('src');
            result.view_url = $el.attr('view');
            result.acquisition_method = this._resolve_acquisition_method($el);
            result.interval = result.acquisition_method ? parseInt($el.attr(result.acquisition_method)) : NaN;

            if($el.attr('enhance')) result.enhance = $el.attr('enhance');

            console.debug('src='+result.src);
            console.debug('view='+result.view_url);
            console.debug('acquisition_method='+result.acquisition_method);
            console.debug('interval='+result.interval);
            console.debug('enhance='+result.enhance);

            return result;
        },
        _check_readonly: function(is_readonly, view_url) {
            if (is_readonly && MVC.String.ends_with(view_url, '_rw')) {
                console.warn('Readonly attribute! Replacing rw view with ro...');
                view_url = view_url.replace('_rw', '_ro');
            }
            return view_url;
        },
        /**
         *
         * @param {*} proxy
         * @param {string} name
         * @param {Array} errors
         * @return {mtango.DeviceAttribute|mtango.DeviceCommand}
         * @private
         */
        _create_wrapped:function(proxy, name, errors) {
            if(!proxy.Class.attributes[name]) {
                errors.push('Device['+proxy.deviceProxy.device+'] has no attribute[' + name + ']');
                return null;
            }
            return proxy['create_new_DeviceAttribute_' + name]();
        },
        /**
         * Creates an instance of this class from jQuery reference
         *
         * Parses $el for required attributes (invokes this._parse)
         *
         * Validates instance created from parsed attributes
         *
         * Creates DeviceAttribute(Command) instance and injects it into created instance
         *
         * Sets $el.id to instance.element_id()
         *
         * @param {jQuery} $el -- $('mtango:attr,mtango:cmd,mtango:img')
         * @return {this}
         */
        from_html:function($el){
            var attributes = this._parse($el);

            var instance = new this(attributes);

            instance.validate();

            //if resulting instance is not valid - return it ui will be updated with instance.errors
            if(!instance.valid()) return instance;

            var device = instance.src.substring(0, instance.src.lastIndexOf('/'));
            var attrName = instance.src.substring(instance.src.lastIndexOf('/') + 1);

            console.debug('device='+device);
            console.debug('attrName='+attrName);
            console.debug('Done parsing.');

            //create DeviceAttribute
            var deviceProxy = mtango.DeviceProxy.find_one(device);
            if(!deviceProxy || !deviceProxy.wrapped_by) {
                instance.errors.push('There is no proxy for device[' + device + ']. Please create one before initializing ' + instance.src);
                return instance;
            }

        //create wrapped
        instance.wrapped = this._create_wrapped(deviceProxy.wrapped_by, attrName, instance.errors);

        //deal with view
            try {
                instance.view = this._resole_view_url_to_view(instance.view_url);
            } catch (e) {
                instance.errors.push(e.message);
            }

            $el.attr('id',instance.element_id());
            return instance;
        }
    },
    {
        $element       : null,
        wrapped: null,
        view           : null,
        /**
         * This value is updated in open() and used in execute()
         *
         */
        argin: null,
        /**
         *
         * @param {jQuery} v
         */
        set_$element       :function(v){
            this.$element = v;
        },
        /**
         *
         * @param {mtango.DeviceAttribute} v
         */
        set_wrapped:function(v){
            this.wrapped = v;
        },
        /**
         *
         * @param {MVC.View} v
         */
        set_view:function(v){
            this.view = v;
        },
        /**
         *
         * @param attributes
         * @constructor
         */
        init:function(attributes){
            if(!attributes.id) attributes.id = this.Class._ID++;
            this._super(attributes);
        },
        _callback:function(){
            return MVC.Function.bind(function(response){
                this.wrapped.response = response;
                this.update();
            },this);
        },
        /**
         * Starts to acquire data
         *
         */
        show:function(){
            if (!this.valid()) {
                this.$element.html(new View({url: mtango.ui.ERROR_VIEW_URL}).render({
                    message: 'Failed to initialize ' + this.$element.prop('tagName'),
                    data   : this
                }));
            } else {
                this.$element.html(new View({url:mtango.ui.LOADING_VIEW_URL}).render());
                this._show();
            }
        },
        /**
         * Internal show
         *
         * @private
         */
        _show:function(){
            this[this.acquisition_method]();
        },
        //TODO hide, destroy ?
        /**
         * Reads from wrapped and updates UI
         *
         */
        read:function(){
            this.wrapped.read(this._callback());
        },
        /**
         * Starts polling
         */
        poll:function(){
            this.wrapped.start_polling(this._callback(),this.interval);
        },
        /**
         * Prompts for a value and writes it
         */
        open:function(){
            var dlgId = '#pre-rendered_' + this.element_id();
            $(dlgId,this.$element).popup();
        },
        /**
         * Popups this spectrum's list dialog
         *
         */
        spectrum_popup:function(){
            $('#pre-rendered_items_' + this.element_id(), this.$element).popup();
        },
        /**
         * Closes associated with this DeviceAttributeUI popup
         */
        close_spectrum_popup:function(ndx){
            this.argin = ndx;
            var dlgId = '#pre-rendered_items_' + this.element_id();
            $(dlgId,this.$element).popup('close');
        },
        /**
         *
         */
        write:function(){
            //TODO loading
            this.argin = $('#dlg_write_arg_' + this.element_id(),this.$element).val();
            //TODO validate argument
            this.wrapped.write(this.argin,
                MVC.Function.bind(function(response){
                    //TODO validate response ???
                    //TODO move read to server
                    this.read();
                },this));
        },
        /**
         *
         */
        change:function(){
            if(this.interval) this.wrapped.timeout = this.interval;
            this.wrapped.change(this._callback());
        },
        /**
         *
         */
        periodic:function(){
            if(this.interval) this.wrapped.timeout = this.interval;
            this.wrapped.periodic(this._callback());
        },
        /**
         *
         */
        archive:function(){
            if(this.interval) this.wrapped.timeout = this.interval;
            this.wrapped.archive(this._callback());
        },
        /**
         *
         */
        user:function(){
            if(this.interval) this.wrapped.timeout = this.interval;
            this.wrapped.user(this._callback());
        },
        /**
         * Creates or gets html representation of this instance (wrapped by jQuery)
         *
         * @return html wrapped into jQuery
         */
        create_element:function(){
            if(this.$element) return this.$element;
            var html = ["<mtango:attr id='",this.element_id(),"' src='",this.src,"' view='", this.view_url,"' ",this.acquisition_method,"='",this.interval,"'><\/mtango:attr>"].join('');
            return this.$element = $(html);
        },
        /**
         * Updates this UI element
         */
        update:function(){
            if(!this.$element) throw "Can not update non existing UI! Try to call #create_element first!";
            if(!this.wrapped.response.valid()) {
                this.$element.html(new View({url:mtango.ui.ERROR_VIEW_URL}).render({
                    message:'Request has failed',
                    data: this.wrapped.response
                }));
            } else {
                document.getElementById(this.element_id()).innerHTML = this.view.render(this);
                if(this.enhance) this.$element.enhanceWithin();
            }
        },
        /**
         * Pauses any activity performed by this ui element
         */
        pause:function(){
            if(this.wrapped){
                this.wrapped.pause_polling();
                this.wrapped.cancel_all();
            }
        },
        /**
         * Resumes activities of this ui element
         */
        resume:function(){
            if(this.wrapped) {
                this.wrapped.resume_polling();
                this._show();
            }
        }

    }
);

if(typeof DeviceAttributeUI == 'undefined') DeviceAttributeUI = mtango.ui.DeviceAttributeUI;